from rich import print

from kubr.backends.base import JobOperationStatus
from kubr.commands.base import BaseCommand
from kubr.commands.utils.reply import confirmation_prompt, mascot_message


class RmCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers, completer):
        rm_parser = subparsers.add_parser("rm", help="Delete a job")
        rm_parser.add_argument("job", help="Name of job to delete").completer = completer
        rm_parser.add_argument("-n", "--namespace", help="Namespace to delete job from", default="default")
        return rm_parser

    def __call__(self, job_name: str, namespace: str = "default"):
        # TODO [rm] add regular expression matching for job names
        # TODO [rm] add completion for namespace typing
        try:
            if confirmation_prompt(f"Are you sure you want to delete job {job_name}?"):
                status = self.backend.delete_job(job_name=job_name, namespace=namespace)
        except Exception as e:
            print(e)
            print(mascot_message(f"Job {job_name} deletion failed!"))
            return

        if status == JobOperationStatus.Success:
            print(mascot_message(f"Job {job_name} was deleted successfully!"))



import pytest
from pydantic_yaml import parse_yaml_raw_as

from kubr.backends.base import BaseBackend
from kubr.backends.volcano import VolcanoBackend
from kubr.config.runner import ExperimentConfig, RunnerConfig

base_config = """
container:
    image: "jannnash/noop:latest"

resources:
    cpu: 1
    mem: 1Gi

experiment:
    name: "pytest-test"
    namespace: "default"
"""


class TestRunner:
    @pytest.fixture
    def runner(self):
        backend = VolcanoBackend()
        runner_cfg = parse_yaml_raw_as(RunnerConfig, base_config)
        return backend, runner_cfg

    def test_run(self, backend: BaseBackend, runner: RunnerConfig):
        backend.run_job(runner)

    @pytest.mark.parametrize(
        "name, namespace, queue",
        [
            ("pytest-test", "default", "default"),
            pytest.param("pytest-test", "default", "test", marks=pytest.mark.xfail),
            pytest.param("pytest-test", "test", "default", marks=pytest.mark.xfail),
            pytest.param("pytest-test", "test", "test", marks=pytest.mark.xfail),
        ],
    )
    def test_experiment_base(
        self, backend: BaseBackend, runner: RunnerConfig, name: str, namespace: str, queue: str, env: dict
    ):
        experiment_config = ExperimentConfig(name=name, namespace=namespace, queue=queue)
        runner.experiment = experiment_config

        backend.run_job(runner)



from typing import List

import cowsay
from rich import print
from rich.table import Table

from kubr.config.job import Job, JobState

mascot = r"""
\
 \
  ╱|、
(˚ˎ 。7
 |、˜〵
じしˍ,)ノ
"""


def mascot_message(msg: str):
    return cowsay.draw(msg, mascot, to_console=False)


def generate_jobs_table(jobs: List[Job], state: str):
    total_gpus = sum([job.gpu for job in jobs])
    show_footer = state in [str(JobState.Running), str(JobState.Pending)]

    table = Table(title=str(state), width=100, show_footer=show_footer, footer_style="bold")
    table.add_column("Name", style="cyan", no_wrap=True, width=60)
    table.add_column("Namespace", style="magenta", justify="center")
    table.add_column("Age", "Total:", style="yellow", justify="center")
    table.add_column("GPU", f"{total_gpus}", style="red", justify="center")
    for job in jobs:
        table.add_row(job.name, job.namespace, job.age, str(job.gpu))
    return table


def confirmation_prompt(msg: str):
    print(mascot_message(msg + "\n |y/N| Default=No"))
    response = input().lower()
    return response in ["y", "yes"]



from enum import Enum


class PrettyEnum(Enum):
    def __str__(self):
        return str(self.value)


class JobOperationStatus(PrettyEnum):
    """DeleteStatus is the status of the job deletion.

    Args:
        Success (str): Job was deleted successfully.
        Failed (str): Job deletion failed.
    """

    Success = "Success"
    Failed = "Failed"


class BaseBackend:
    def __init__(self):
        pass

    def run_job(self, *args, **kwargs):
        raise NotImplementedError

    def list_jobs(self, *args, **kwargs):
        raise NotImplementedError

    def delete_jobs(self, *args, **kwargs):
        raise NotImplementedError

    def get_logs(self, *args, **kwargs):
        raise NotImplementedError

    def describe_job(self, *args, **kwargs):
        raise NotImplementedError



from datetime import datetime
from time import sleep
from typing import Optional

import humanize
from kubernetes import watch
from pydantic_yaml import parse_yaml_raw_as
from rich import print
from rich.console import Console, Group
from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress

from kubr.backends.base import JobOperationStatus
from kubr.commands.base import BaseCommand
from kubr.commands.utils.reply import confirmation_prompt, generate_jobs_table, mascot_message
from kubr.config.job import Job, JobState
from kubr.config.runner import RunnerConfig


def visualize_job(job):
    job.age = humanize.naturaldelta(datetime.utcnow() - job.age)
    print(generate_jobs_table([job], state=str(JobState.Pending)))


class RunCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers):
        run_parser = subparsers.add_parser("run", help="Submit a new job")
        run_parser.add_argument("config", help="Path to run config", type=str)
        run_parser.add_argument("-i", "--image", help="Image to run")
        run_parser.add_argument("-e", "--entrypoint", help="Entrypoint to run")
        run_parser.add_argument("-n", "--namespace", help="Namespace to submit job to")
        run_parser.add_argument("--name", help="Name of job")
        run_parser.add_argument("-v", "--verbose", help="Verbose output", action="store_true", default=False)

    def show_job_run(self, job: Job):
        node_update_step = 100 / job.nodes
        console = Console()

        w = watch.Watch()
        stream = w.stream(
            self.backend.core_client.list_namespaced_event,
            namespace=job.namespace,
        )
        autoscaler = []

        progress = Progress()
        status = console.status("Running job...")

        live_panel = Live(Panel(Group(status, progress)))
        live_panel.start()

        scheduling_pb = progress.add_task("[red]Scheduling...", total=100)
        init_pb = progress.add_task("[green]Init...", total=100)
        container_pb = progress.add_task("[cyan]Container...", total=100)

        run_state = "Scheduling"

        pod_events = []
        started_containers = 0

        # involvedObject.name={job.name}*,
        for line in stream:
            event_object = line["object"]
            if not event_object.metadata.name.startswith(job.name):
                continue
            if event_object.source.component == "cluster-autoscaler":
                autoscaler.append(event_object)
                continue

            if event_object.reason == "FailedScheduling":
                pod_events.append(event_object)

            if event_object.message.startswith("Successfully assigned"):
                progress.update(scheduling_pb, advance=node_update_step)
                if progress.tasks[scheduling_pb].completed:
                    run_state = "Init"

            if event_object.message == f"Started container {job.name}-init":
                progress.update(init_pb, advance=node_update_step)
                if progress.tasks[scheduling_pb].completed:
                    run_state = "Container"

            if event_object.message == f"Started container {job.name}":
                progress.update(container_pb, advance=node_update_step)
                started_containers += 1
                if started_containers == job.nodes:
                    status.update("Waiting for logs...")
                    break

            if "pod group is not ready" in event_object.message and len(pod_events):
                status.update(f"{run_state}... {event_object.message}\n \t Pod issue: {pod_events[-1].message}")
            else:
                status.update(f"{run_state}... {event_object.message}")

        log_found = 0
        while True:
            sleep(2)
            try:
                log_stream = self.backend.get_logs(job_name=job.name, namespace=job.namespace, tail=None, follow=True)
                for log in log_stream:
                    # TODO [run] stop if log stopped
                    if not log_found:
                        status.update("Job started!")
                        live_panel.stop()
                    print(log)
                    log_found += 1

                break
            except Exception:
                status.update("Waiting for logs...")

    def __call__(
        self,
        config: str,
        image: Optional[str] = None,
        name: Optional[str] = None,
        entrypoint: Optional[str] = None,
        namespace: Optional[str] = None,
        verbose: bool = False,
    ):
        # TODO [run] check if config exists on cluster and ask to resubmit

        with open(config, "r") as f:
            config = f.read()
        config = parse_yaml_raw_as(RunnerConfig, config)

        config.experiment.name = name or config.experiment.name
        config.container.image = image or config.container.image
        config.container.entrypoint = entrypoint or config.container.entrypoint
        config.experiment.namespace = namespace or config.experiment.namespace

        if "-" in config.experiment.name:
            print(mascot_message(f"Job name {config.experiment.name} is invalid!"))
            return

        pods = self.backend.core_client.list_namespaced_pod(
            namespace=config.experiment.namespace, label_selector=f"volcano.sh/job-name={config.experiment.name}"
        )
        if len(pods.items) > 0:
            if confirmation_prompt(
                f"Job {config.experiment.name} already exists in namespace {config.experiment.namespace}. \n"
                f"Do you want to resubmit it?"
            ):
                self.backend.delete_job(job_name=config.experiment.name, namespace=config.experiment.namespace)

        job, status = self.backend.run_job(config)
        if status == JobOperationStatus.Failed:
            print(mascot_message(f"Job {config.experiment.name} running failed!"))

        elif status == JobOperationStatus.Success:
            if verbose:
                self.show_job_run(job)
            else:
                visualize_job(job)
        else:
            print(mascot_message(f"Job {config.experiment.name} is in unknown state!"))



# PYTHON_ARGCOMPLETE_OK
import argparse

import argcomplete

from kubr.backends.volcano import VolcanoBackend
from kubr.commands.logs import LogsCommand
from kubr.commands.ls import LsCommand
from kubr.commands.rm import RmCommand
from kubr.commands.run import RunCommand


def main():
    # TODO fix package install adding eval in bash\zsh for autocomplete
    # adding eval "$(register-python-argcomplete kubr)" to .bashrc or .zshrc
    # TODO fix autopilot deployment in GKE autopilot
    # TODO fix Volcano priority class in GKE (https://github.com/volcano-sh/volcano/issues/2379)

    backend = VolcanoBackend()
    arg = argparse.ArgumentParser(description="Kubr", add_help=True)
    arg.add_argument("--version", help="Get version of Kubr")
    subparsers = arg.add_subparsers(help="Commands", dest="command")

    RunCommand.add_parser(subparsers)
    LsCommand.add_parser(subparsers)
    RmCommand.add_parser(subparsers, completer=backend._completion_list_running_jobs)
    LogsCommand.add_parser(subparsers, completer=backend._completion_list_running_jobs)

    # DescribeCommand.add_parser(subparsers, completer=backend._completion_list_running_jobs)
    # attach_parser = AttachCommand.add_parser(subparsers, completer=backend._completion_list_running_jobs)
    # stat_parser = StatCommand.add_parser(subparsers, completer=backend._completion_list_running_jobs)
    # test_parser = TestCommand.add_parser(subparsers)

    argcomplete.autocomplete(arg)
    args = arg.parse_args()

    if args.command == "run":
        operator = RunCommand()
        operator(
            config=args.config,
            image=args.image,
            entrypoint=args.entrypoint,
            namespace=args.namespace,
            name=args.name,
            verbose=args.verbose,
        )
    elif args.command == "ls":
        operator = LsCommand()
        operator(namespace=args.namespace, show_all=args.all, head=args.top)
    elif args.command == "rm":
        operator = RmCommand()
        operator(job_name=args.job, namespace=args.namespace)
    # elif args.command == "desc":
    #     operator = DescribeCommand()
    #     operator(job_name=args.job_name, namespace=args.namespace)
    elif args.command == "logs":
        operator = LogsCommand()
        operator(job_name=args.job, namespace=args.namespace, tail=args.tail, follow=args.follow)
    elif args.command == "attach":
        raise NotImplementedError  # TODO implement attach command
    elif args.command == "stat":
        raise NotImplementedError  # TODO implement stat command
    elif args.command == "test":
        raise NotImplementedError  # TODO implement test command -- run IB\scheduler\metrics\registry\ethernet tests
    else:
        arg.print_help()


if __name__ == "__main__":
    main()



from kubr.backends.volcano import VolcanoBackend


class BaseCommand:
    def __init__(self, backend=None):
        self.backend = backend or VolcanoBackend()

    @staticmethod
    def add_parser(subparsers, completer=None):
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        raise NotImplementedError



from collections import defaultdict
from datetime import datetime
from typing import List, Optional

import humanize
from rich import print
from rich.columns import Columns

from kubr.commands.base import BaseCommand
from kubr.commands.utils.reply import generate_jobs_table, mascot_message
from kubr.config.job import Job, JobState


def visualize_jobs(jobs: List[Job], head: Optional[int] = 10, show_all: bool = False):
    extracted_jobs = defaultdict(list)
    for job in jobs:
        if job.state in [JobState.Pending, JobState.Running, JobState.Completed, JobState.Failed]:
            extracted_jobs[str(job.state)].append(job)
        else:
            extracted_jobs["extra"].append(job)

    for state in ["Pending", "Running", "Completed", "Failed", "extra"]:
        extracted_jobs[state].sort(key=lambda x: x.age, reverse=True)
        for job in extracted_jobs[state]:
            job.age = humanize.naturaldelta(datetime.utcnow() - job.age)
        if head:
            extracted_jobs[state] = extracted_jobs[state][:head]

        # TODO [ls] add pretty formatting that will show only first 10 jobs in completed and failed states are shown
        if not show_all and state in ["Completed", "Failed"]:
            extracted_jobs[state] = extracted_jobs[state][:5]

        if len(extracted_jobs[state]):
            extracted_jobs[state] = generate_jobs_table(jobs=extracted_jobs[state], state=state)
        else:
            extracted_jobs[state] = mascot_message(f"No {state} jobs found!")

    result = [
        extracted_jobs["Running"],
        extracted_jobs["Pending"],
        extracted_jobs["Completed"],
        extracted_jobs["Failed"],
    ]

    if show_all:
        result.append(extracted_jobs["extra"])

    return Columns(
        result,
        equal=True,
        align="center",
    )


class LsCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers):
        ls_parser = subparsers.add_parser("ls", help="List all jobs")
        ls_parser.add_argument("-n", "--namespace", help="Namespace to list jobs from", default="All")
        ls_parser.add_argument("-a", "--all", help="Show all jobs", action="store_true", default=False)
        ls_parser.add_argument("-t", "--top", help="Show only first T jobs", default=None, type=int)
        return ls_parser

    def __call__(self, namespace: str = "All", head: Optional[int] = None, show_all: bool = False):
        jobs = self.backend.list_jobs(namespace=namespace)
        print(visualize_jobs(jobs=jobs, head=head, show_all=show_all))



from kubr.commands.base import BaseCommand


class TestCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers, completer):
        test_parser = subparsers.add_parser("test", help="Test a job")
        return test_parser

    def __call__(
        self,
    ):
        raise NotImplementedError



import re
from datetime import datetime
from enum import Enum
from typing import Any, Dict, Iterable, Mapping, Optional

import humanize
from kubernetes import client, config, watch
from rich import print
from tabulate import tabulate

from kubr.backends.base import BaseBackend, JobOperationStatus
from kubr.backends.k8s_runner import create_pod_definition
from kubr.config.job import Job, JobBackend, JobState, JobType
from kubr.config.runner import EnvVar, RunnerConfig


def normalize_str(data: str) -> str:
    """
    Invokes ``lower`` on thes string and removes all
    characters that do not satisfy ``[a-z0-9\\-]`` pattern.
    This method is mostly used to make sure kubernetes and gcp_batch scheduler gets
    the job name that does not violate its restrictions.
    """
    if data.startswith("-"):
        data = data[1:]
    pattern = r"[a-z0-9\-]"
    return "".join(re.findall(pattern, data.lower()))


class RetryPolicy(str, Enum):
    REPLICA = "REPLICA"
    APPLICATION = "APPLICATION"


RETRY_POLICIES: Mapping[str, Iterable[Mapping[str, str]]] = {
    RetryPolicy.REPLICA: [],
    RetryPolicy.APPLICATION: [
        {"event": "PodEvicted", "action": "RestartJob"},
        {"event": "PodFailed", "action": "RestartJob"},
    ],
}


class VolcanoBackend(BaseBackend):
    DEFAULT_TASK_NAME = "worker"

    def __init__(self):
        self.kubernetes_config = config.load_config()
        self.crd_client = client.CustomObjectsApi()
        self.core_client = client.CoreV1Api()

    def run_job(self, run_config: RunnerConfig) -> [Job, JobOperationStatus]:
        tasks = []

        for replica_id in range(run_config.resources.nodes):
            rank0_env = f"VC_{normalize_str(self.DEFAULT_TASK_NAME)}_0_HOSTS".upper()
            if replica_id == 0:
                rank0_env = "KUBR_RANK0_HOST"
                run_config.container.env.append(EnvVar(name="KUBR_RANK0_HOST", value="localhost"))

            pod = create_pod_definition(
                pod_name=run_config.experiment.name,
                runner_config=run_config,
                service_account=None,
                rank0_env=rank0_env,
            )

            # pod.metadata.labels.update(
            #     pod_labels(
            #         app=app,
            #         role_idx=role_idx,
            #         role=role,
            #         replica_id=replica_id,
            #         app_id=unique_app_id,
            #     )
            # )

            task: Dict[str, Any] = {
                "replicas": 1,
                "name": f"{self.DEFAULT_TASK_NAME}-{replica_id}",
                "template": pod,
            }
            if run_config.experiment.worker_max_retries > 0:
                task["maxRetry"] = run_config.experiment.worker_max_retries
                task["policies"] = RETRY_POLICIES[RetryPolicy.APPLICATION]

            task["minAvailable"] = 1

            tasks.append(task)

        job_spec = {
            "schedulerName": "volcano",
            "queue": run_config.experiment.queue,
            "tasks": tasks,
            "maxRetry": run_config.experiment.job_retries,
            "plugins": {
                # https://github.com/volcano-sh/volcano/issues/533
                "svc": ["--publish-not-ready-addresses"],
                "env": [],
            },
        }
        # if run_config.experiment.priority_class is not None:
        #     job_spec["priorityClassName"] = run_config.experiment.priority_class

        resource: Dict[str, object] = {
            "apiVersion": "batch.volcano.sh/v1alpha1",
            "kind": "Job",
            "metadata": {"name": f"{run_config.experiment.name}"},
            "spec": job_spec,
        }

        try:
            self.crd_client.create_namespaced_custom_object(
                group="batch.volcano.sh",
                version="v1alpha1",
                namespace=run_config.experiment.namespace,
                plural="jobs",
                body=resource,
            )

        except Exception as e:
            # TODO [run] add exception printing
            print(e)
            return None, JobOperationStatus.Failed

        job = Job(
            type=JobType.torchrun,
            backend=JobBackend.Volcano,
            name=run_config.experiment.name,
            namespace=run_config.experiment.namespace,
            state=JobState.Pending,
            age=datetime.now(),
            gpu=run_config.resources.gpu * run_config.resources.nodes,
            nodes=run_config.resources.nodes,
        )
        return job, JobOperationStatus.Success

    def _completion_list_running_jobs(self, **kwargs):
        print("Using completion list for running jobs")
        jobs_stat = self.crd_client.list_cluster_custom_object(
            group="batch.volcano.sh", version="v1alpha1", plural="jobs"
        )
        jobs = jobs_stat["items"]
        running_jobs = []
        for job in jobs:
            if job["status"]["state"]["phase"] == "Running":
                running_jobs.append(job["metadata"]["name"])
        return running_jobs

    @staticmethod
    def _extract_gpu_count(k8s_job) -> int:
        gpu_count = 0

        for task in k8s_job["spec"]["tasks"]:
            for container in task["template"]["spec"]["containers"]:
                if "nvidia.com/gpu" in container["resources"]["limits"]:
                    gpu_count += int(container["resources"]["limits"]["nvidia.com/gpu"])
        return gpu_count

    def list_jobs(self, namespace: str = "All"):
        # TODO [ls] speedup for selected namespace(filter on server side)
        jobs_stat = self.crd_client.list_cluster_custom_object(
            group="batch.volcano.sh", version="v1alpha1", plural="jobs"
        )
        k8s_jobs = jobs_stat["items"]

        extracted_jobs = []
        for k8s_job in k8s_jobs:
            if namespace != "All" and k8s_job["metadata"]["namespace"] != namespace:
                continue

            job = Job(
                type=JobType.torchrun,
                backend=JobBackend.Volcano,
                name=k8s_job["metadata"]["name"],
                namespace=k8s_job["metadata"]["namespace"],
                state=k8s_job["status"]["state"]["phase"],
                age=datetime.strptime(k8s_job["status"]["state"]["lastTransitionTime"], "%Y-%m-%dT%H:%M:%SZ"),
                gpu=self._extract_gpu_count(k8s_job),
            )
            extracted_jobs.append(job)

        return extracted_jobs

    def delete_job(self, job_name: str, namespace: str) -> JobOperationStatus:
        # TODO add cli response formatting for deletion confirmation
        self.crd_client.delete_namespaced_custom_object(
            group="batch.volcano.sh", version="v1alpha1", namespace=namespace, plural="jobs", name=job_name
        )
        events = self.core_client.list_namespaced_event(namespace=namespace)
        for event in events.items:
            if event.metadata.name.startswith(f"{job_name}"):
                self.core_client.delete_namespaced_event(name=event.metadata.name, namespace=namespace)

        return JobOperationStatus.Success

    def get_job_main_pod(self, job_name: str, namespace: str):
        pods = self.core_client.list_namespaced_pod(
            namespace=namespace, label_selector=f"volcano.sh/job-name={job_name}"
        )
        # TODO add logic for multi pod master-worker selection for logging extraction
        if len(pods.items) == 0:
            raise Exception(f"No pods found for job {job_name} in namespace {namespace}")
        pod = pods.items[0]
        pod_name = pod.metadata.name
        return pod_name, pod

    def get_job_events(self, pod_name: str, namespace: str):
        events = self.core_client.list_namespaced_event(
            namespace=namespace, field_selector=f"involvedObject.name={pod_name}"
        )
        return events

    def get_logs(self, job_name: str, namespace: str, tail: Optional[int] = None, follow: bool = False):
        pod_name, pod = self.get_job_main_pod(job_name, namespace)
        containers = pod.spec.containers
        if len(containers) == 0:
            return f"No containers found for pod {pod_name} in namespace {namespace}"
        # TODO add logic for multi container selection for logging extraction
        container = containers[0]
        container_name = container.name
        if tail:
            return self.core_client.read_namespaced_pod_log(
                name=pod_name, namespace=namespace, container=container_name, tail_lines=tail
            )
        if follow:
            w = watch.Watch()
            return w.stream(
                self.core_client.read_namespaced_pod_log, name=pod_name, namespace=namespace, container=container_name
            )

        api_response = self.core_client.read_namespaced_pod_log(name=pod_name, namespace=namespace)
        return api_response

    def describe_job(self, job_name: str, namespace: str):
        pod_name, pod = self.get_job_main_pod(job_name, namespace)
        raw_events = self.get_job_events(pod_name, namespace)

        events = []
        for event in raw_events.items:
            events.append(
                {
                    "Last Seen": event.last_timestamp,
                    "From": event.source.component,
                    "Type": event.type,
                    "Reason": event.reason,
                    "Message": event.message,
                }
            )
        events.sort(key=lambda x: x["Last Seen"], reverse=True)
        for event in events:
            event["Last Seen"] = humanize.naturaldelta(datetime.now() - event["Last Seen"].replace(tzinfo=None))

        # TODO add handling events longer than 10
        events = events[:10]

        events = tabulate(events, headers="keys", tablefmt="grid", maxcolwidths=80)

        return events



from kubr.commands.base import BaseCommand


class StatCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers, completer):
        stat_parser = subparsers.add_parser("stat", help="Get statistics about a job")
        stat_parser.add_argument("job", help="Name of job to get statistics about").completer = completer
        stat_parser.add_argument("-n", "--namespace", help="Namespace to get statistics from", default="default")
        return stat_parser

    def __call__(self, job_name: str, namespace: str = "default"):
        raise NotImplementedError



from typing import List, Literal, Optional, Union

import pydantic
from pydantic import BaseModel

from kubr.config.job import JobBackend, JobType


class SecretMount(BaseModel):
    env: str
    secret_namespace: str
    secret_name: str
    secret_key: str


class EnvVar(BaseModel):
    """EnvVar is the configuration for an environment variable.

    Args:
        name (str): Name of the environment variable.
        value (str): Value of the environment variable.
    """

    name: str
    value: str


class ContainerConfig(BaseModel):
    """ContainerConfig is the configuration for the container.

    Args:
        image (str): Image to run.
        entrypoint (Optional[str], optional): Entrypoint to run. Defaults to None.
        env (Dict[str, str], optional): Environment variables to pass to the entrypoint. Defaults to {}.
        secrets (Optional[List[SecretConfig]], optional): Secrets to pass to the entrypoint. Defaults to None.
    """

    image: str
    entrypoint: Optional[str] = None
    env: List[EnvVar] = []
    secrets: List[SecretMount] = []
    # port_map: Dict[str, int] = field(default_factory=dict)
    # args: List[str] = field(default_factory=list)
    # python_path: str = ""


class GitConfig(BaseModel):
    url: str
    branch: str
    commit: str


class VolumeMount(BaseModel):
    """VolumeMount is the configuration for a volume mount.

    Args:
        name (str): Name of the volume.
        type (Literal["hostPath"]): Type of the volume.
        mount_path (str): Mount path of the volume.
    """

    name: str
    type: Literal["hostPath"]
    mount_path: str


class CodePersistenceConfig(BaseModel):
    # git: Optional[GitConfig] = None
    # pvc: Optional[str] = None
    volume: Optional[VolumeMount] = None


class DataConfig(BaseModel):
    """DataConfig is the configuration for the data.

    Args:
        volumes (Optional[List[VolumeMount]], optional): List of volumes to mount. Defaults to [].
    """

    # pvcs: Optional[List[str]] = None
    volumes: Optional[List[VolumeMount]] = []


class ResourceConfig(BaseModel):
    """ResourceConfig is the configuration for the resources.

    Args:
        nodes (int, optional): Number of replicas to run. Defaults to 1.
        cpu (int, optional): Number of CPUs to request. Defaults to 0.
        memory (int, optional): Memory in GB to request. Defaults to 0.
        gpu (int, optional): Number of GPUs to request. Defaults to 0.
        ib (Union[int, Literal['auto']], optional): Number of Infiniband devices to request. Defaults to 0.
        ib_device (str, optional): Name of the Infiniband device to request. Defaults to "nvidia.com/hostdev".
    """

    # TODO [config][resources] add taints\tolerations\affinity
    nodes: int = pydantic.Field(gt=0, type=int, default=1)
    cpu: int = 0
    memory: float = 0
    gpu: int = 0
    # devices: Dict[str, float] = field(default_factory=dict)
    # capabilities: Dict[str, str] = field(default_factory=dict)
    ib: Union[int, Literal["auto"]] = 0
    ib_device: str = "nvidia.com/hostdev"


class ExperimentConfig(BaseModel):
    """ExperimentConfig is the configuration for the experiment.

    Args:
        name (str): Name of the experiment.
        namespace (str): Namespace of the experiment.
        queue (Optional[str], optional): Queue to submit the experiment to. Defaults to "default".
        job_retries (int, optional): Number of retries for the job. Defaults to 0.
        worker_max_retries (int, optional): Maximum number of retries for the task. Defaults to 10.
    """

    name: str
    namespace: str

    # args: List[str] = field(default_factory=list)
    # env: Dict[str, str] = field(default_factory=dict)

    queue: Optional[str] = "default"
    # priority_class: Optional[str] = None

    # TODO add tests for retries
    job_retries: int = 0
    worker_max_retries: int = 0


class RunnerConfig(BaseModel):
    """RunnerConfig is the configuration for the runner.

    Args:
        experiment (ExperimentConfig): Experiment configuration.
        container (ContainerConfig): Container configuration.
        resources (ResourceConfig): Resource configuration.
        type (JobType, optional): Job type. Defaults to JobType.torchrun.
        backend (JobBackend, optional): Job backend. Defaults to JobBackend.Volcano.
        code (Optional[CodePersistenceConfig], optional): Code persistence configuration. Defaults to None.
        data (Optional[DataConfig], optional): Data configuration. Defaults to None.

    """

    experiment: ExperimentConfig
    init_container: Optional[ContainerConfig] = None
    container: ContainerConfig
    resources: ResourceConfig
    type: JobType = JobType.torchrun
    backend: JobBackend = JobBackend.Volcano
    code: Optional[CodePersistenceConfig] = None
    data: Optional[DataConfig] = None



from kubr.commands.base import BaseCommand


class AttachCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers, completer):
        attach_parser = subparsers.add_parser("attach", help="Attach to a running job")
        attach_parser.add_argument("job", help="Name of job to attach to").completer = completer
        attach_parser.add_argument("-n", "--namespace", help="Namespace to attach to", default="default")
        return attach_parser

    def __call__(self, job_name: str, namespace: str = "default"):
        raise NotImplementedError



from rich import print

from kubr.commands.base import BaseCommand


class DescribeCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers, completer):
        desc_parser = subparsers.add_parser("desc", help="Get info about a job")
        desc_parser.add_argument("job_name", help="Name of job to get info about").completer = completer
        desc_parser.add_argument("-n", "--namespace", help="Namespace to get info from", default="default")
        return desc_parser

    def __call__(self, *args, **kwargs):
        result = self.backend.describe_job(*args, **kwargs)
        print(result)



import datetime
from typing import Union

from pydantic import BaseModel

from kubr.backends.base import PrettyEnum


class JobType(PrettyEnum):
    """JobType is the type of job to run.

    Args:
        torchrun (str): Run a torchrun job.
    """

    torchrun = "torchrun"
    basic = "basic"


class JobBackend(PrettyEnum):
    """JobBackend is the type of job to run.

    Args:
        Volcano (str): Run a Volcano job.
    """

    Volcano = "Volcano"


class JobState(PrettyEnum):
    """JobState is the state of the job.

    Args:
        Pending (str): Job is pending.
        Running (str): Job is running.
        Completed (str): Job is completed.
        Failed (str): Job has failed.
    """

    Pending = "Pending"
    Running = "Running"
    Completed = "Completed"
    Failed = "Failed"


class Job(BaseModel):
    type: JobType
    backend: JobBackend
    name: str
    namespace: str
    state: JobState
    age: Union[datetime.datetime, str]
    gpu: int
    nodes: int = 1



import pydoc
from typing import Optional

from rich import print

from kubr.backends.volcano import VolcanoBackend
from kubr.commands.utils.reply import mascot_message


class LogsCommand:
    def __init__(self, backend=None):
        self.backend = backend or VolcanoBackend()

    @staticmethod
    def add_parser(subparsers, completer):
        logs_parser = subparsers.add_parser("logs", help="Get logs of a job")
        logs_parser.add_argument("job", help="Name of job to get logs of").completer = completer
        logs_parser.add_argument("-n", "--namespace", help="Namespace to get logs from", default="default")
        logs_parser.add_argument("-t", "--tail", help="Number of lines to show", default=None, type=int)
        logs_parser.add_argument("-f", "--follow", help="Follow logs", action="store_true", default=False)
        return logs_parser

    def __call__(self, job_name: str, namespace: str, tail: Optional[int] = None, follow: bool = False):
        try:
            logs = self.backend.get_logs(job_name=job_name, namespace=namespace, tail=tail, follow=follow)
        except Exception as e:
            print(e)
            print(mascot_message(f"Job {job_name} logs retrieval failed!"))
            return
        if follow:
            # TODO [logs][follow] add pretty Ctrl+C handling
            for log in logs:
                print(log)
        else:
            pydoc.pager(logs)



import shlex
from typing import Dict, Iterable, Optional

from kubernetes.client import V1ContainerPort, V1EnvVarSource, V1HostPathVolumeSource, V1SecretKeySelector
from kubernetes.client.models import (  # noqa: F811 redefinition of unused
    V1Container,
    V1EmptyDirVolumeSource,
    V1EnvVar,
    V1ObjectMeta,
    V1Pod,
    V1PodSpec,
    V1ResourceRequirements,
    V1SecurityContext,
    V1Volume,
    V1VolumeMount,
)

from kubr.config.job import JobType
from kubr.config.runner import ContainerConfig, DataConfig, ResourceConfig, RunnerConfig

RESERVED_MILLICPU = 100
RESERVED_MEMMB = 1024

ANNOTATION_ISTIO_SIDECAR = "sidecar.istio.io/inject"


class _noquote(str):
    pass


def _args_join(args: Iterable[str]) -> str:
    """
    _args_join is like shlex.join but if the argument is wrapped in _noquote
    it'll not quote that argument.
    """
    quoted = [arg if isinstance(arg, _noquote) else shlex.quote(arg) for arg in args]
    return " ".join(quoted)


def create_pod_definition(
    pod_name: str,
    runner_config: RunnerConfig,
    service_account: Optional[str],
    rank0_env: str,
) -> "V1Pod":
    resource_config: ResourceConfig = runner_config.resources
    container_config: ContainerConfig = runner_config.container
    data_config: Optional[DataConfig] = runner_config.data
    init_container_config: Optional[ContainerConfig] = runner_config.init_container

    rdzv_port: int = 29500

    limits = {}
    requests = {}

    if resource_config.cpu > 0:
        mcpu = int(resource_config.cpu * 1000)
        limits["cpu"] = f"{mcpu}m"
        request_mcpu = max(mcpu - RESERVED_MILLICPU, 0)
        requests["cpu"] = f"{request_mcpu}m"
    if resource_config.memory > 0:
        limits["memory"] = f"{int(resource_config.memory * 2 ** 10)}M"
        request_memMB = max(int(resource_config.memory) - RESERVED_MEMMB, 0)
        requests["memory"] = f"{request_memMB}M"
    if resource_config.gpu > 0:
        requests["nvidia.com/gpu"] = limits["nvidia.com/gpu"] = str(resource_config.gpu)
    if resource_config.ib > 0:
        requests[resource_config.ib_device] = limits[resource_config.ib_device] = str(resource_config.ib)

    # for device_name, device_limit in resource_config.devices.items():
    #     limits[device_name] = str(device_limit)

    resources = V1ResourceRequirements(
        limits=limits,
        requests=requests,
    )

    node_selector: Dict[str, str] = {}
    # if LABEL_INSTANCE_TYPE in resource.capabilities:
    #     node_selector[LABEL_INSTANCE_TYPE] = resource.capabilities[LABEL_INSTANCE_TYPE]

    SHM_VOL = "dshm"
    volumes = [
        V1Volume(
            name=SHM_VOL,
            empty_dir=V1EmptyDirVolumeSource(
                medium="Memory",
            ),
        ),
    ]
    volume_mounts = [
        V1VolumeMount(name=SHM_VOL, mount_path="/dev/shm"),
    ]
    if data_config is not None:
        for volume in data_config.volumes:
            if volume.type == "hostPath":
                host_path, mount_path = volume.mount_path.split(":")
                volumes.append(
                    V1Volume(
                        name=volume.name,
                        host_path=V1HostPathVolumeSource(
                            path=host_path,
                        ),
                    )
                )
                volume_mounts.append(
                    V1VolumeMount(
                        name=volume.name,
                        mount_path=mount_path,
                    )
                )
            else:
                raise ValueError(f"Unknown volume type {volume.type}")
    security_context = V1SecurityContext()

    container_envs = []
    for env in container_config.env:
        container_envs.append(
            V1EnvVar(
                name=env.name,
                value=env.value,
            )
        )
    for secret in container_config.secrets:
        container_envs.append(
            V1EnvVar(
                name=secret.env,
                value_from=V1EnvVarSource(
                    secret_key_ref=V1SecretKeySelector(
                        name=secret.secret_name,
                        key=secret.secret_key,
                    ),
                ),
            )
        )

    # port_maps = [
    #     V1ContainerPort(
    #         name=name,
    #         container_port=port,
    #     )
    #     for name, port in container_config.port_map.items()
    # ]
    port_maps = []
    if runner_config.type == JobType.torchrun:
        port_maps.append(
            V1ContainerPort(
                name="c10d",
                container_port=rdzv_port,
            )
        )
    init_containers = []
    if init_container_config is not None:
        # TODO [run] support env handling for init_container
        init_containers.append(
            V1Container(
                command=init_container_config.entrypoint.split()
                if init_container_config.entrypoint is not None
                else None,
                image=init_container_config.image,
                image_pull_policy="Always",
                name=f"{pod_name}-init",
            )
        )
    cmd = []

    if runner_config.type == JobType.torchrun:
        rdzv_backend = "c10d"
        # rank0_env = "${rank0_env}"
        if runner_config.resources.nodes > 1:
            rdzv_endpoint = _noquote(f"$${{{rank0_env}:=localhost}}:{rdzv_port}")
        else:
            rdzv_endpoint = "localhost:0"

        cmd = [
            "torchrun",
            "--rdzv_backend",
            rdzv_backend,
            "--rdzv_endpoint",
            rdzv_endpoint,
            "--rdzv_id",
            f"{runner_config.experiment.name}",
            "--rdzv_conf",
            "read_timeout=600",
            "--nnodes",
            str(runner_config.resources.nodes),
            "--nproc_per_node",
            str(runner_config.resources.gpu),
            "--tee",
            "3",
            "--role",
            "",
        ]
    elif runner_config.type == JobType.basic:
        cmd = []
    else:
        raise ValueError(f"Unknown job type {runner_config.type}")

    cmd += container_config.entrypoint.split() if container_config.entrypoint is not None else []

    if runner_config.type == JobType.torchrun:
        cmd = ["bash", "-c", _args_join(cmd)]

    container = V1Container(
        command=cmd,
        image=container_config.image,
        image_pull_policy="Always",
        name=pod_name,
        env=container_envs,
        resources=resources,
        ports=port_maps,
        volume_mounts=volume_mounts,
        security_context=security_context,
    )

    return V1Pod(
        spec=V1PodSpec(
            init_containers=init_containers,
            containers=[container],
            restart_policy="Never",
            service_account_name=service_account,
            volumes=volumes,
            node_selector=node_selector,
        ),
        metadata=V1ObjectMeta(
            annotations={
                # Disable the istio sidecar as it prevents the containers from
                # exiting once finished.
                ANNOTATION_ISTIO_SIDECAR: "false",
            },
            labels={},
        ),
    )
